# -*- coding: utf8 -*-
# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

from __future__ import annotations

__all__ = ['upload_project']

import logging
import os
import shutil
import tempfile
import typing

from binstar_client import errors

if typing.TYPE_CHECKING:
    import argparse
    import types
    import anaconda_project.project
    import anaconda_project.internal.simple_status


logger = logging.getLogger('binstar.projects.upload')


class UploadedProject(typing.TypedDict):
    """Details on uploaded project."""

    username: str
    name: str
    url: str


class _TmpDir:

    def __init__(self, prefix: str = '') -> None:
        self._dir: str = tempfile.mkdtemp(prefix=prefix)

    def __enter__(self) -> str:
        return self._dir

    def __exit__(
            self,
            _type: typing.Optional[typing.Type[Exception]] = None,
            value: typing.Optional[Exception] = None,
            traceback: typing.Optional[types.TracebackType] = None,
    ) -> None:
        try:
            shutil.rmtree(path=self._dir)
        except Exception as error:  # pylint: disable=broad-except
            logger.debug('Failed to clean up TmpDir "%s": %s', self._dir, str(error))


def _real_upload_project(
        project: anaconda_project.project.Project,
        args: argparse.Namespace,
        username: str
) -> UploadedProject:
    try:
        # pylint: disable=import-outside-toplevel
        from anaconda_project import project_ops
    except ImportError as error:
        raise errors.BinstarError('anaconda-project package is not installed') from error

    logger.info('Uploading project: %s', project.name)
    status: anaconda_project.internal.simple_status.SimpleStatus = project_ops.upload(
        project,
        site=args.site,
        username=username,
        token=args.token,
        log_level=args.log_level,
    )
    if status:
        logger.info(status.status_description)
        return {
            'username': username,
            'name': project.name,
            'url': status.url,
        }

    status_error: typing.Any
    for status_error in status.errors:
        logger.error(str(status_error))
    logger.error(status.status_description)
    raise errors.BinstarError(status.status_description)


def upload_project(project_path: str, args: argparse.Namespace, username: str) -> UploadedProject:
    try:
        # pylint: disable=import-outside-toplevel
        from anaconda_project import project
        from anaconda_project import project_ops
    except ImportError as error:
        message: str = f'To upload projects such as "{project_path}", install the anaconda-project package.'
        logger.error(message)
        raise errors.BinstarError(message) from error

    if os.path.exists(project_path) and os.path.isdir(project_path):
        new_project = project.Project(directory_path=project_path)
        return _real_upload_project(new_project, args, username)

    # make the single file into a temporary project directory
    with (_TmpDir(prefix='anaconda_upload_')) as dirname:
        shutil.copy(project_path, dirname)
        basename_no_extension = os.path.splitext(os.path.basename(project_path))[0]
        new_project = project_ops.create(dirname, name=basename_no_extension)
        return _real_upload_project(new_project, args, username)



# -*- coding: utf-8 -*-

# pylint: disable=redefined-outer-name,import-outside-toplevel,line-too-long
# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

"""Utilities for determining application-specific dirs."""

import os.path


class EnvAppDirs:

    def __init__(self, root_path):
        self.root_path = root_path

    @property
    def user_data_dir(self):
        return os.path.join(self.root_path, 'data')

    @property
    def user_config_dir(self):
        return os.path.join(self.root_path, 'data')

    @property
    def site_data_dir(self):
        return os.path.join(self.root_path, 'data')

    @property
    def user_cache_dir(self):
        return os.path.join(self.root_path, 'cache')

    @property
    def user_log_dir(self):
        return os.path.join(self.root_path, 'log')



# -*- coding: utf8 -*-

"""Package type detection and meta-data extraction."""

from __future__ import annotations

__all__ = ['detect_package_meta', 'detect_package_extension', 'detect_package_type', 'get_attrs']

import collections

import functools

import logging
import tarfile
import typing

from os import path

from binstar_client.utils.config import PackageType
from binstar_client.inspect_package.conda import inspect_conda_package
from binstar_client.inspect_package import conda_installer
from binstar_client.inspect_package.pypi import inspect_pypi_package
from binstar_client.inspect_package.r import inspect_r_package
from binstar_client.inspect_package.ipynb import inspect_ipynb_package
from binstar_client.inspect_package.env import inspect_env_package

if typing.TYPE_CHECKING:
    import typing_extensions

    P = typing_extensions.ParamSpec('P')


logger = logging.getLogger('binstar.detect')


class Meta(typing.NamedTuple):
    """General details on detected package."""

    package_type: PackageType
    extension: str


Raw: typing_extensions.TypeAlias = typing.Optional[str]
OptMeta: typing_extensions.TypeAlias = typing.Optional[Meta]


def checker_for(package_type: PackageType) -> typing.Callable[[typing.Callable[P, Raw]], typing.Callable[P, OptMeta]]:
    """Convert function to a checker for a particular :code:`package_type`."""
    def wrapper(function: typing.Callable[P, Raw]) -> typing.Callable[P, OptMeta]:
        @functools.wraps(function)
        def wrapped(*args: P.args, **kwargs: P.kwargs) -> OptMeta:
            extension: typing.Optional[str] = function(*args, **kwargs)
            if extension is None:
                return None
            return Meta(package_type=package_type, extension=extension)
        return wrapped
    return wrapper


RawDetector: typing_extensions.TypeAlias = typing.Callable[[str], Raw]
Detector: typing_extensions.TypeAlias = typing.Callable[[str], OptMeta]
DETECTORS: typing.Final[typing.OrderedDict[PackageType, Detector]] = collections.OrderedDict()


def detector_for(package_type: PackageType) -> typing.Callable[[RawDetector], Detector]:
    """
    Same as :func:`~checker_for`, but also registers checker in the registry of detectors.

    .. warning::

        All detectors must be defined in the order they must be used to check for a package type.
    """
    def register(function: RawDetector) -> Detector:
        result: Detector = checker_for(package_type)(function)
        if DETECTORS.setdefault(package_type, result) is not result:
            raise ValueError(f'detector for "{package_type.value}" is already registered')
        return result
    return register


def find_postfix(value: str, *options: str) -> typing.Optional[str]:
    """Find a first applicable postfix for a :code:`value` from available :code:`options`."""
    option: str
    for option in options:
        if value.endswith(option):
            return option
    return None


@detector_for(PackageType.CONDA)
def is_conda(filename: str) -> typing.Optional[str]:
    """
    Check if :code:`filename` is a conda package.

    :return: File extension if it is a conda package, otherwise - :code:`None`.
    """
    logger.debug('Testing if %s is a conda package ..', filename)
    if filename.endswith('.conda'):
        return '.conda'
    if filename.endswith('.tar.bz2'):  # Could be a conda package
        try:
            tar_file: tarfile.TarFile
            with tarfile.open(filename, mode='r:bz2') as tar_file:
                tar_file.getmember('info/index.json')
        except KeyError:
            logger.debug("Not a conda package: no 'info/index.json' file in the tarball")
            return None
        logger.debug('This is a conda package')
        return '.tar.bz2'
    logger.debug('Not a conda package (should be a .tar.bz2 or .conda file)')
    return None


@detector_for(PackageType.STANDARD_PYTHON)
def is_pypi(filename: str) -> typing.Optional[str]:
    """
    Check if :code:`filename` is a python package.

    :return: File extension if it is a python package, otherwise - :code:`None`.
    """
    package_type_label: str = PackageType.STANDARD_PYTHON.label
    logger.debug('Testing if %s is a %s package ..', filename, package_type_label)
    if filename.endswith('.whl'):
        logger.debug('This is a %s wheel package', package_type_label)
        return '.whl'
    result: typing.Optional[str] = find_postfix(filename, '.tar.gz', 'tgz')
    if result is None:
        logger.debug('This is not a %s package (expected .tgz, .tar.gz or .whl)', package_type_label)
        return None
    tar_file: tarfile.TarFile
    with tarfile.open(filename) as tar_file:  # Could be a setuptools sdist or r source package
        if any(name.endswith('/PKG-INFO') for name in tar_file.getnames()):
            return result
    logger.debug("This is not a %s package (no '/PKG-INFO' in tarball)", package_type_label)
    return None


@detector_for(PackageType.STANDARD_R)
def is_r(filename: str) -> typing.Optional[str]:
    """
    Check if :code:`filename` is an r package.

    :return: File extension if it is an r package, otherwise - :code:`None`.
    """
    logger.debug('Testing if %s is an R package ..', filename)
    result: typing.Optional[str] = find_postfix(filename, '.tar.gz', 'tgz')
    if result is None:
        logger.debug('This not is an R package (expected .tgz, .tar.gz).')
        return None
    tar_file: tarfile.TarFile
    with tarfile.open(filename) as tar_file:  # Could be a setuptools sdist or r source package
        name: str
        flags: int = 0
        for name in tar_file.getnames():
            if name.endswith('/DESCRIPTION'):
                flags |= 1
            elif name.endswith('/NAMESPACE'):
                flags |= 2
            else:
                continue
            if flags == 3:
                return result
    logger.debug("This not is an R package (no '*/DESCRIPTION' and '*/NAMESPACE' files).")
    return None


@detector_for(PackageType.NOTEBOOK)
def is_ipynb(filename: str) -> typing.Optional[str]:
    """
    Check if :code:`filename` is a notebook.

    :return: File extension if it is a notebook, otherwise - :code:`None`.
    """
    logger.debug('Testing if %s is an ipynb file ..', filename)
    result: typing.Optional[str] = find_postfix(filename, '.ipynb')
    if result is None:
        logger.debug('Not an ipynb file')
    return result


@detector_for(PackageType.ENV)
def is_environment(filename: str) -> typing.Optional[str]:
    """
    Check if :code:`filename` is an environment.

    :return: File extension if it is an environment, otherwise - :code:`None`.
    """
    logger.debug('Testing if %s is an environment file ..', filename)
    result: typing.Optional[str] = find_postfix(filename, '.yml', '.yaml')
    if result is None:
        logger.debug('Not an environment file')
    return result


is_installer: typing.Final[Detector] = (  # pylint: disable=invalid-name
    detector_for(PackageType.INSTALLER)(conda_installer.is_installer)
)


@detector_for(PackageType.PROJECT)
def is_project(filename: str) -> typing.Optional[str]:
    """
    Check if :code:`filename` is a project.

    :return: File extension if it is a project, otherwise - :code:`None`.
    """
    logger.debug('Testing if %s is a project ..', filename)
    if path.isdir(filename):
        return ''
    result: typing.Optional[str] = find_postfix(filename, '.py')
    if result is None:
        logger.debug('Not a project')
    return result


def complete_package_meta(filename: typing.Union[str, bytes], package_type: PackageType) -> OptMeta:
    """Collect package metadata on a :code:`filename` with known :code:`package_type`."""
    if isinstance(filename, bytes):
        filename = filename.decode('utf-8', errors='ignore')

    detector: typing.Optional[Detector] = DETECTORS.get(package_type, None)
    return detector and detector(filename)


def detect_package_meta(filename: typing.Union[str, bytes]) -> OptMeta:
    """Detect package type of a :code:`filename` with additional metadata on it."""
    if isinstance(filename, bytes):
        filename = filename.decode('utf-8', errors='ignore')

    detector: Detector
    result: OptMeta
    for detector in DETECTORS.values():
        if result := detector(filename):
            return result

    return None


def detect_package_extension(filename: typing.Union[str, bytes]) -> typing.Optional[str]:
    """Detect an extension of a package located at :code:`filename`."""
    result: OptMeta = detect_package_meta(filename)
    return result and result.extension


def detect_package_type(filename: typing.Union[str, bytes]) -> typing.Optional[PackageType]:
    """Detect a package type of a :code:`filename`."""
    result: OptMeta = detect_package_meta(filename)
    return result and result.package_type


# ======================================================================================================================

PackageAttributes: typing_extensions.TypeAlias = typing.Dict[str, typing.Any]
ReleaseAttributes: typing_extensions.TypeAlias = typing.Dict[str, typing.Any]
FileAttributes: typing_extensions.TypeAlias = typing.Dict[str, typing.Any]
Attributes: typing_extensions.TypeAlias = typing.Tuple[PackageAttributes, ReleaseAttributes, FileAttributes]


class Inspector(typing.Protocol):  # pylint: disable=too-few-public-methods
    """Common interface for package inspectors."""

    def __call__(self, filename: str, fileobj: typing.BinaryIO, *args: typing.Any, **kwargs: typing.Any) -> Attributes:
        """Collect metadata on a package."""


def inspect_file(filename, fileobj, *args, **kwargs):  # pylint: disable=unused-argument
    """Collect metadata on a generic file."""
    return {}, {'description': ''}, {'basename': path.basename(filename), 'attrs': {}}


INSPECTORS: typing.Final[typing.Mapping[PackageType, Inspector]] = {
    PackageType.CONDA: inspect_conda_package,
    PackageType.STANDARD_PYTHON: inspect_pypi_package,
    PackageType.STANDARD_R: inspect_r_package,
    PackageType.NOTEBOOK: inspect_ipynb_package,
    PackageType.ENV: inspect_env_package,
    PackageType.INSTALLER: conda_installer.inspect_package,
    PackageType.FILE: inspect_file,
}


def get_attrs(package_type: PackageType, filename: str, *args: typing.Any, **kwargs: typing.Any) -> Attributes:
    """Collect metadata attributes on a package."""
    fileobj: typing.BinaryIO
    with open(filename, 'rb') as fileobj:
        return INSPECTORS[package_type](filename, fileobj, *args, **kwargs)



"""Multipart form utils."""
import typing

import requests
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor

if typing.TYPE_CHECKING:
    import tqdm


def multipart_files_upload(
        url: str,
        data: typing.MutableMapping,
        files: typing.Optional[typing.Mapping[str, tuple]] = None,
        progress_bar: typing.Optional['tqdm.tqdm'] = None,
        **request_kwargs: typing.Any) -> requests.Response:
    """
    Uploads one or more files as a multipart form.

    :param url: The URL to which the files will be uploaded.
    :param data: Dictionary, list of tuples, bytes, or file-like object to send in as a multipart form.
    :param files: Dictionary of ``{'name': file-tuple}`` for multipart encoding upload.
    :param progress_bar: An optional progress bar to display the upload progress.
    :param request_kwargs: Any additional keyword arguments to pass to the `requests.post()` function.

    """
    if files:
        data.update(files)

    encoder = MultipartEncoder(data)

    if progress_bar:
        encoder = MultipartEncoderMonitor(
            encoder, lambda monitor: progress_bar.update(monitor.bytes_read - progress_bar.n)
        )

    return requests.post(
        url,
        data=encoder,
        headers={'Content-Type': encoder.content_type},
        timeout=request_kwargs.pop('timeout', 10 * 60 * 60),
        **request_kwargs,
    )
